#include "multypo.h"

#include <QX11Info>
#include <QDebug>
#include <QVBoxLayout>
#include <QLabel>
#include <QMap>

#include <X11/Xlib.h>
#include <X11/extensions/XInput2.h>
#include <X11/Xutil.h>

MultypoWindow::MultypoWindow(QWidget *parent) :
	QWidget(parent),
	xiInited(false),
	gameStarted(false)
{
	/* Prepare colors */
	setStyleSheet("background-color: black; color: green; font-size: 45pt");

	/* Prepare conents */
	setLayout(new QVBoxLayout(this));

	mainLabel = new QLabel("Namen eingeben, dann Return", this);
	layout()->addWidget(mainLabel);

	/* Fullscreen, no cursor */
	setWindowState(Qt::WindowFullScreen);
	setCursor(QCursor(Qt::BlankCursor));

	words.open(stdin,QIODevice::ReadOnly);

	typingPlayers = 0;
}

MultypoWindow::~MultypoWindow()
{
}

void MultypoWindow::resetPlayerText() {
	for (QMap<int, Player*>::Iterator it = players.begin(); it != players.end(); ++it) {
		it.value()->resetText();
	}
}

void MultypoWindow::nextWord() {
	QByteArray tmp = words.readLine().trimmed();
	QString word = QString::fromUtf8(tmp);
	qDebug() << "New word" << word;
	if (word.isEmpty()) { // game over
		mainLabel->setText("GAME OVER");
		for (QMap<int, Player*>::Iterator it = players.begin(); it != players.end(); ++it) {
			QString tmp = QString ("Spieler %1 hat %2 Punkte.").arg(it.value()->getName()).arg(it.value()->score);
			it.value()->setWaiting(tmp);
		}
	} else {
		resetPlayerText();
		mainLabel->setText(word);
		typingPlayers = players.size();
	}
}

void MultypoWindow::handleKeyPress(int device, QString string)
{
	qDebug() << "Device" << device << "String" << string;

	if (string == "Escape") {
		close();
		return;
	}

	if (!players.contains(device)) {
		if (gameStarted)
			return;
		players[device] = new Player(this);
	}
	bool newChar = players[device]->handleKey(string);

	if (gameStarted) { // ingame
		qDebug() << "current player line" << players[device]->getCurrentLine();
		qDebug() << "current word" << mainLabel->text();
		if (newChar && players[device]->getCurrentLine() == mainLabel->text()) {
			players[device]->score += typingPlayers;
			QString readyString = QString("READY: %1 points").arg(typingPlayers);
			players[device]->setWaiting(readyString);
			qDebug() << "typingPlayers " << typingPlayers;
			typingPlayers--;
			qDebug() << "typingPlayers " << typingPlayers;
			if (typingPlayers <= 0) {
				nextWord();
			}
		}
	} else { // name entering phase
		qDebug() << "checking for game started";
		bool allHaveNames = true;
		for (QMap<int, Player*>::Iterator it = players.begin(); it != players.end(); ++it) {
			if (! it.value()->hasName()) {
				allHaveNames = false;
				break;
			}
		}
		qDebug() << "Players empty?" << players.empty() << "All have names?" << allHaveNames;
		if (!players.empty() && allHaveNames) {
			gameStarted = true;
			nextWord();
		}
	}
}

bool MultypoWindow::handleX11Event(XEvent *event)
{
	Display *dpy = x11Info().display();
	/* Handle the first map event: We are finally visible! */
	if (!xiInited && event->type == MapNotify) {
		/* XInput Extension available? */
		int eventID, errorID;
		if (!XQueryExtension(dpy, "XInputExtension", &xiOpcode, &eventID, &errorID)) {
		   qCritical() << "X Input extension not available";
		}

		/* Which version of XI2? We support 2.0 */
		int major = 2, minor = 1;
		if (XIQueryVersion(dpy, &major, &minor) == BadRequest) {
			qCritical() << "XI2 not available. Server supports" << major << "." <<
						minor;
		}

		qDebug() << "System supports XI2";


		// Now let's listen to keypress events
		XIEventMask eventmask;
		unsigned char mask [1] = { 0 }; // will change
		eventmask.deviceid = XIAllMasterDevices;
		eventmask.mask_len = sizeof (mask); // in bytes, for whatever reason...
		eventmask.mask = mask;
		XISetMask(mask, XI_KeyPress);
		XISelectEvents (dpy, winId(), &eventmask, 1);

		// finally, grab all focuses
		int ndevices;
		XIDeviceInfo* devices = XIQueryDevice(dpy, XIAllMasterDevices, &ndevices);
		for (int i = 0; i < ndevices; ++i) {
			if (devices[i].use != XIMasterKeyboard) continue;
			qDebug() << "Found master keyboard with ID" << devices[i].deviceid;
			XISetFocus(dpy, devices[i].deviceid, winId(), CurrentTime);
		}
		XIFreeDeviceInfo(devices);

		xiInited = true;

		return false;
	}
	else if (xiInited && event->type == GenericEvent
		&& event->xcookie.type == GenericEvent && event->xcookie.extension == xiOpcode
		&& event->xcookie.evtype == XI_KeyPress) { 
		
		if (XGetEventData(dpy, &event->xcookie) != True) {
			qCritical() << "Error getting event data";
			// FIXME return true;
		}
		
		/* Handle XI event */
		XIDeviceEvent *d_ev = (XIDeviceEvent*) event->xcookie.data;
	    KeyCode keycode = d_ev->detail;
	    int kbdid = d_ev->deviceid;

		if (!(d_ev->flags & XIKeyRepeat)) {
			int keysyms_per_keycode;
			KeySym keysym = XGetKeyboardMapping (dpy, keycode, 1, &keysyms_per_keycode)[0];
			handleKeyPress(kbdid, XKeysymToString(keysym));
		}

		XFreeEventData(dpy, &event->xcookie);

		return true;
	}
	else {
		return false;
	}
}
